from resotolib.utils import num_default_threads
from resotolib.config import Config
from resoto_plugin_aws import AWSCollectorPlugin


def test_args():
    config = Config("dummy", "dummy")
    AWSCollectorPlugin.add_config(config)
    Config.init_default_config()
    assert Config.aws.access_key_id is None
    assert Config.aws.secret_access_key is None
    assert Config.aws.role is None
    assert Config.aws.role_override is False
    assert Config.aws.account is None
    assert Config.aws.region is None
    assert Config.aws.scrape_org is False
    assert Config.aws.fork_process is True
    assert Config.aws.scrape_exclude_account == []
    assert Config.aws.assume_current is False
    assert Config.aws.do_not_scrape_current is False
    assert Config.aws.account_pool_size == num_default_threads()
    assert Config.aws.region_pool_size == 20
    assert Config.aws.parallel_api_requests == 10
    assert len(Config.aws.collect) == 0
    assert len(Config.aws.no_collect) == 0



import botocore.exceptions
import multiprocessing
import resotolib.proc
import resotolib.logger
from resotolib.logger import log, setup_logger
from concurrent import futures
from resotolib.args import ArgumentParser
from argparse import Namespace
from resotolib.config import Config, RunningConfig
from resotolib.graph import Graph
from resotolib.utils import log_runtime
from resotolib.baseplugin import BaseCollectorPlugin
from .config import AwsConfig
from .utils import aws_session
from .resources import AWSAccount
from .accountcollector import AWSAccountCollector
from prometheus_client import Summary, Counter
from typing import List


resotolib.logger.getLogger("boto").setLevel(resotolib.logger.CRITICAL)

metrics_collect = Summary(
    "resoto_plugin_aws_collect_seconds", "Time it took the collect() method"
)
metrics_unhandled_account_exceptions = Counter(
    "resoto_plugin_aws_unhandled_account_exceptions_total",
    "Unhandled AWS Plugin Account Exceptions",
    ["account"],
)


class AWSCollectorPlugin(BaseCollectorPlugin):
    cloud = "aws"

    def __init__(self) -> None:
        super().__init__()
        self.__regions = []

    @staticmethod
    def add_config(config: Config) -> None:
        config.add_config(AwsConfig)

    @metrics_collect.time()
    def collect(self) -> None:
        log.debug("plugin: AWS collecting resources")
        if not self.authenticated:
            log.error("Failed to authenticate - skipping collection")
            return

        if Config.aws.assume_current and not Config.aws.do_not_scrape_current:
            log.warning(
                "You specified assume_current but not do_not_scrape_current! "
                "This will result in the same account being scraped twice and is likely not what you want."
            )

        if Config.aws.role and Config.aws.scrape_org:
            accounts = [
                AWSAccount(aws_account_id, {}, role=Config.aws.role)
                for aws_account_id in get_org_accounts(
                    filter_current_account=not Config.aws.assume_current
                )
                if aws_account_id not in Config.aws.scrape_exclude_account
            ]
            if not Config.aws.do_not_scrape_current:
                accounts.append(AWSAccount(current_account_id(), {}))
        elif Config.aws.role and Config.aws.account:
            accounts = [
                AWSAccount(aws_account_id, {}, role=Config.aws.role)
                for aws_account_id in Config.aws.account
            ]
        else:
            accounts = [AWSAccount(current_account_id(), {})]

        max_workers = (
            len(accounts)
            if len(accounts) < Config.aws.account_pool_size
            else Config.aws.account_pool_size
        )
        pool_args = {"max_workers": max_workers}
        if Config.aws.fork_process:
            pool_args["mp_context"] = multiprocessing.get_context("spawn")
            pool_args["initializer"] = resotolib.proc.initializer
            pool_executor = futures.ProcessPoolExecutor
        else:
            pool_executor = futures.ThreadPoolExecutor

        with pool_executor(**pool_args) as executor:
            wait_for = [
                executor.submit(
                    collect_account,
                    account,
                    self.regions,
                    ArgumentParser.args,
                    Config.running_config,
                )
                for account in accounts
            ]
            for future in futures.as_completed(wait_for):
                account_graph = future.result()
                if not isinstance(account_graph, Graph):
                    log.error(
                        f"Returned account graph has invalid type {type(account_graph)}"
                    )
                    continue
                self.graph.merge(account_graph)

    @property
    def regions(self) -> List:
        if len(self.__regions) == 0:
            if not Config.aws.region or (
                isinstance(Config.aws.region, list) and len(Config.aws.region) == 0
            ):
                log.debug("AWS region not specified, assuming all regions")
                self.__regions = all_regions()
            else:
                self.__regions = list(Config.aws.region)
        return self.__regions

    @property
    def authenticated(self) -> bool:
        try:
            _ = current_account_id()
        except botocore.exceptions.NoCredentialsError:
            log.error("No AWS credentials found")
            return False
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "AuthFailure":
                log.error("AWS was unable to validate the provided access credentials")
            elif e.response["Error"]["Code"] == "InvalidClientTokenId":
                log.error("AWS was unable to validate the provided security token")
            elif e.response["Error"]["Code"] == "ExpiredToken":
                log.error("AWS security token included in the request is expired")
            else:
                raise
            return False
        return True


def current_account_id():
    session = aws_session()
    return session.client("sts").get_caller_identity().get("Account")


def get_org_accounts(filter_current_account=False):
    session = aws_session()
    client = session.client("organizations")
    accounts = []
    try:
        response = client.list_accounts()
        accounts = response.get("Accounts", [])
        while response.get("NextToken") is not None:
            response = client.list_accounts(NextToken=response["NextToken"])
            accounts.extend(response.get("Accounts", []))
    except botocore.exceptions.ClientError as e:
        if e.response["Error"]["Code"] == "AccessDeniedException":
            log.error("AWS error - missing permissions to list organization accounts")
        else:
            raise
    filter_account_id = current_account_id() if filter_current_account else -1
    accounts = [
        aws_account["Id"]
        for aws_account in accounts
        if aws_account["Id"] != filter_account_id
    ]
    for account in accounts:
        log.debug(f"AWS found org account {account}")
    log.info(f"AWS found a total of {len(accounts)} org accounts")
    return accounts


def all_regions() -> List:
    session = aws_session()
    ec2 = session.client("ec2", region_name="us-east-1")
    regions = ec2.describe_regions()
    return [r["RegionName"] for r in regions["Regions"]]


@log_runtime
def collect_account(
    account: AWSAccount,
    regions: List,
    args: Namespace = None,
    running_config: RunningConfig = None,
) -> Graph:
    collector_name = f"aws_{account.id}"
    resotolib.proc.set_thread_name(collector_name)

    if args is not None:
        ArgumentParser.args = args
        setup_logger("resotoworker-aws")
    if running_config is not None:
        Config.running_config.apply(running_config)

    log.debug(f"Starting new collect process for account {account.dname}")

    aac = AWSAccountCollector(regions, account)
    try:
        aac.collect()
    except botocore.exceptions.ClientError as e:
        log.exception(
            f"An AWS {e.response['Error']['Code']} error occurred while collecting account {account.dname}"
        )
        metrics_unhandled_account_exceptions.labels(account=account.dname).inc()
    except Exception:
        log.exception(
            f"An unhandled error occurred while collecting AWS account {account.dname}"
        )
        metrics_unhandled_account_exceptions.labels(account=account.dname).inc()

    return aac.graph



from resotolib.utils import num_default_threads
from dataclasses import dataclass, field
from typing import List, ClassVar, Optional


@dataclass
class AwsConfig:
    kind: ClassVar[str] = "aws"
    access_key_id: Optional[str] = field(
        default=None,
        metadata={
            "description": "AWS Access Key ID (null to load from env - recommended)"
        },
    )
    secret_access_key: Optional[str] = field(
        default=None,
        metadata={
            "description": "AWS Secret Access Key (null to load from env - recommended)"
        },
    )
    role: Optional[str] = field(
        default=None, metadata={"description": "IAM role name to assume"}
    )
    role_override: bool = field(
        default=False,
        metadata={
            "description": "Override any stored role names (e.g. from remote graphs)"
        },
    )
    account: Optional[List[str]] = field(
        default=None,
        metadata={
            "description": "List of AWS Account ID(s) to collect (null for all if scrape_org is true)"
        },
    )
    region: Optional[List[str]] = field(
        default=None,
        metadata={"description": "List of AWS Regions to collect (null for all)"},
    )
    scrape_org: bool = field(
        default=False, metadata={"description": "Scrape the entire AWS organization"}
    )
    fork_process: bool = field(
        default=True,
        metadata={"description": "Fork collector process instead of using threads"},
    )
    scrape_exclude_account: List[str] = field(
        default_factory=list,
        metadata={"description": "List of accounts to exclude when scraping the org"},
    )
    assume_current: bool = field(
        default=False, metadata={"description": "Assume given role in current account"}
    )
    do_not_scrape_current: bool = field(
        default=False, metadata={"description": "Do not scrape current account"}
    )
    account_pool_size: int = field(
        default_factory=num_default_threads,
        metadata={"description": "Account thread/process pool size"},
    )
    region_pool_size: int = field(
        default=20, metadata={"description": "Region thread pool size"}
    )
    parallel_api_requests: int = field(
        default=10,
        metadata={
            "description": "Maximum number of parallel API requests per account/region"
        },
    )
    collect: List[str] = field(
        default_factory=list,
        metadata={"description": "List of AWS services to collect (default: all)"},
    )
    no_collect: List[str] = field(
        default_factory=list,
        metadata={"description": "List of AWS services to exclude (default: none)"},
    )



import boto3
import boto3.session
import uuid
from typing import Iterable, List, Dict
from resotolib.config import Config
from resotolib.baseresources import BaseRegion, BaseResource
from resotolib.graph import Graph
from retrying import retry
from prometheus_client import Counter
from botocore.exceptions import ConnectionClosedError, CredentialRetrievalError


metrics_session_exceptions = Counter(
    "resoto_plugin_aws_session_exceptions_total",
    "Unhandled AWS Plugin Session Exceptions",
)


def retry_on_session_error(e):
    if isinstance(e, (ConnectionClosedError, CredentialRetrievalError)):
        metrics_session_exceptions.inc()
        return True
    return False


@retry(
    stop_max_attempt_number=10,
    wait_random_min=1000,
    wait_random_max=6000,
    retry_on_exception=retry_on_session_error,
)
def aws_session(aws_account=None, aws_role=None):
    if Config.aws.role_override:
        aws_role = Config.aws.role
    if aws_role and aws_account:
        role_arn = f"arn:aws:iam::{aws_account}:role/{aws_role}"
        session = boto3.session.Session(
            aws_access_key_id=Config.aws.access_key_id,
            aws_secret_access_key=Config.aws.secret_access_key,
            region_name="us-east-1",
        )
        sts = session.client("sts")
        token = sts.assume_role(
            RoleArn=role_arn, RoleSessionName=f"{aws_account}-{str(uuid.uuid4())}"
        )
        credentials = token["Credentials"]
        return boto3.session.Session(
            aws_access_key_id=credentials["AccessKeyId"],
            aws_secret_access_key=credentials["SecretAccessKey"],
            aws_session_token=credentials["SessionToken"],
        )
    else:
        return boto3.session.Session(
            aws_access_key_id=Config.aws.access_key_id,
            aws_secret_access_key=Config.aws.secret_access_key,
        )


def aws_client(resource: BaseResource, service: str, graph: Graph = None):
    return aws_session(resource.account(graph).id, resource.account(graph).role).client(
        service, region_name=resource.region(graph).id
    )


def aws_resource(resource: BaseResource, service: str, graph: Graph = None):
    return aws_session(
        resource.account(graph).id, resource.account(graph).role
    ).resource(service, region_name=resource.region(graph).id)


def paginate(method: callable, **kwargs) -> Iterable:
    """Get a paginator for a boto3 list/describe method

    Example Usage:
    session = aws_session(self.account.id, self.account.role)
    client = session.client('autoscaling', region_name=region.id)
    for autoscaling_group in paginate(client.describe_auto_scaling_groups):
        print(autoscaling_group)
    """
    client = method.__self__
    paginator = client.get_paginator(method.__name__)
    for page in paginator.paginate(**kwargs).result_key_iters():
        for result in page:
            yield result


def arn_partition(region: BaseRegion):
    arn_partition = "aws"
    if region.id.startswith("cn-"):
        arn_partition = "aws-cn"
    elif region.id.startswith("us-gov-"):
        arn_partition = "aws-us-gov"
    return arn_partition


def tags_as_dict(tags: List) -> Dict:
    return {tag["Key"]: tag["Value"] for tag in tags or []}
